#!/usr/bin/env python

from pathlib import Path
from typing import Any
from urllib.parse import quote as urlquote

import markdown
import yaml

# The paths that are always to be ignored.
STATIC_IGNORE = [
    '.obsidian',
    '.dragonglass'
]

MARKDOWN_PAT = '*.md'


class SourceNode:
    def __init__(self, root: Path, path: Path, is_dir: bool) -> None:
        self._root = root
        self._path = path
        self._is_dir = is_dir
        self._is_md = path.match(MARKDOWN_PAT)
        self.metadata: dict[str, Any] | None = None
        self.text: str | None = None

    def __str__(self) -> str:
        return f"SourceNode({self._path}, {self._is_dir}) [is_md={self._is_md}]"

    @property
    def is_dir(self) -> bool:
        return self._is_dir

    @property
    def is_md(self) -> bool:
        return self._is_md

    @property
    def path(self) -> Path:
        return self._path

    def target_file(self, dest_dir: Path) -> Path:
        p = self._path
        if self._is_md:
            p = p.with_suffix('.html')
        return dest_dir / p

    def link_target(self, prefix: str = "/") -> str:
        xpath = self._path.with_suffix('.html') if self._is_md else self._path
        return urlquote(prefix + xpath.as_posix())

    def load_metadata(self, source_dir: Path) -> None:
        if self._is_md and not self._is_dir:
            with open(source_dir / self._path, "r", encoding="utf-8") as f:
                cur_line = f.readline()
                if cur_line == '---\n':
                    metalines: list[str] = []
                    cur_line = f.readline()
                    while cur_line != '---\n':
                        metalines.append(cur_line)
                        cur_line = f.readline()
                    self.metadata = yaml.full_load(''.join(metalines))

    def parse_markdown(self, source_dir: Path, markdown_parser: markdown.Markdown) -> None:
        if self._is_md and not self._is_dir:
            markdown_parser.reset()
            with open(source_dir / self._path, "r", encoding="utf-8") as f:
                self.text = markdown_parser.convert(f.read())

    def make_vars(self) -> dict[str, Any]:
        return {
            "text": self.text,
            "title": self._path.stem
        }


def generate_list(source_root: Path) -> list[SourceNode]:
    nodes: list[SourceNode] = []
    dirs = [source_root]
    while len(dirs) > 0:
        current_dir = dirs.pop(0)
        for child in current_dir.iterdir():
            rchild = child.relative_to(source_root)
            add_me = True
            for pat in STATIC_IGNORE:
                if rchild.match(pat):
                    add_me = False
                    break
            if add_me:
                nodes.append(SourceNode(source_root, rchild, child.is_dir()))
                if child.is_dir():
                    dirs.append(child)
    return nodes


class SourceIndex:
    def __init__(self, nodelist: list[SourceNode]) -> None:
        self._byname: dict[str, SourceNode] = {}
        self._byalias: dict[str, SourceNode] = {}
        for node in nodelist:
            if node.is_dir:
                continue
            if node.is_md:
                tmp = node.path.with_suffix('')
                key = tmp.name
                if key not in self._byname:
                    self._byname[key] = node
                self._byname[tmp.as_posix()] = node
                if node.metadata:
                    aliases: list[str] = node.metadata.get('aliases', [])
                    for alias in aliases:
                        if alias not in self._byalias:
                            self._byalias[alias] = node
            else:
                key = node.path.name
                if key not in self._byname:
                    self._byname[key] = node
                self._byname[node.path.as_posix()] = node

    def lookup(self, reference: str) -> tuple[SourceNode | None, str | None]:
        if reference in self._byname:
            return self._byname[reference], 'NAME'
        elif reference in self._byalias:
            return self._byalias[reference], 'ALIAS'
        else:
            return None, None
